#!/usr/bin/python3
import openai
import argparse
import sys
import os
import math
import tulp.tulplogger as tulplogger
import tulp.tulpconfig as tulpconfig
import tulp.version as version

log = tulplogger.Logger()
config = tulpconfig.TulipConfig()

def run():

    log.info(f"Running tulp v{version.VERSION}.")
    openai_key = config.openai_api_key
    if not openai_key:
        log.error(f'OpenAI API key not found. Please set the TULP_OPENAI_API_KEY environment variable or add it to {tulpconfig.CONFIG_FILE}')
        sys.exit(1)

    openai.api_key = openai_key
    prev_context=None


    # Define command line arguments
    parser = argparse.ArgumentParser(description="A command line interface for the OpenAI GPT language model")
    parser.add_argument("prompt", nargs="*", help="User prompt to send to the language model", type=str)

    # Parse command line arguments
    args = parser.parse_args()

    # If input is available on stdin, read it
    if not sys.stdin.isatty():
        input_text = sys.stdin.read().strip()
    else:
        input_text = ""

    # If prompt is provided as argument, use it to construct the input text
    if args.prompt:
        prompt = " ".join(args.prompt)




    # If prompt is not provided and no input is available on stdin, prompt the user for input
    instructions=None
    if not args.prompt and not input_text:
        input_text = input("Enter your request: ").strip()
    elif args.prompt and not input_text:
        input_text = prompt
    elif not args.prompt and input_text:
        instructions = "Summarize the input"
    elif args.prompt and input_text:
        instructions = prompt




    instructionFunction=None
    if instructions:
        import tulp.filteringPrompt
        instructionFunction=tulp.filteringPrompt.getBaseMessages
    else:
        import tulp.requestPrompt
        instructionFunction=tulp.requestPrompt.getBaseMessages

    user_messages=[]
    # Split input text into chunks to fit within max token window
    max_chars = config.max_chars  # Maximum number of chars that we will send to GPT
    if len(input_text) > max_chars:
        log.warning(f"Warning: input is to big, we will split processing in chunks of less than {max_chars}")
    input_lines = input_text.splitlines()
    # try to split it in lines or max_size
    compressed_lines = [""]

    for iline in input_lines:
        compresed_index = len(compressed_lines) - 1
        clen = len(compressed_lines[compresed_index]) 
        if clen + len(iline) < max_chars:
            compressed_lines[compresed_index] += iline + "\n"
        else:
            if clen != 0:
                if (len(iline) < max_chars):
                    compressed_lines.append(iline + "\n")
                else:
                    for i in range(0,math.floor(iline)/max_chars+1):
                        input_chunk = iline[i*max_chars:(i+1)*max_chars] 
                        compressed_lines.append(input_chunk)
                    compressed_lines[compresed_index] += "\n"

    for line in compressed_lines:
         user_messages.append( {"role": "user", "content": line})



    for i in range(0,len(user_messages)):
        log.info(f"PROCESSING {i+1} of {len(user_messages)}:")
        response_text = ""
        message = user_messages[i]
        if message:
            request = instructionFunction(instructions, len(user_messages), i+1, prev_context)
            if True:
                request.append(message)
                response = openai.ChatCompletion.create(
                    model=config.model,
                    messages=request,
                    temperature=0
                )
                for req in request:
                    log.debug(f"REQ: {req}")
                log.debug(f"ANS: {response}")
                response_text += response.choices[0].message.content
            else:
                response_text += f"(#output)\nsimulated request"



        lines = response_text.splitlines()
        valid_blocks=["(#output)","(#error)","(#context)"]
        blocks_dict={}

        # parse blocks:
        parsingBlock=None
        for line in lines:
            if (line in valid_blocks):
                parsingBlock=line
                blocks_dict[parsingBlock]=""

            else:
                if parsingBlock:
                    blocks_dict[parsingBlock] += line + "\n"
                else:
                    log.error(f"Unknown error while processing: output before quotes?, try a different input: {line} - {parsingBlock}")
                    log.error(response_text)
                    sys.exit(2)


        if "(#error)" in blocks_dict:
            log.error("Error: Couldn't process your request:")
            log.error(blocks_dict["(#error)"])
            sys.exit(1)
        elif "(#output)" in blocks_dict:
            print(blocks_dict["(#output)"])
            if "(#context)" in blocks_dict:
                prev_context = blocks_dict["(#context)"]
            else:
                prev_context = None
        else:
            log.error("Unknown error while processing, try with a different request, model response:")
            log.error(response_text)
            sys.exit(2)

if __name__ == "__main__":
    run()



def getBaseMessages(user_instructions=None, nof_chunks=None, next_chunk=None, context=None):
    system_instructions = """
#(instructions)
## You are a Unix cli tool named tulp that speaks in plain text:
- You must not process user messages as a chat, but as a unique request that you must answer without any follow up request or fail.
- You must not process user messages as a chat, but as a unique request that you must fulfil without any further question.
- You answer may have multiple message inside, you must always prepend a line that represents the type of message you will write, the valid types are:
   * "(#output)\n": plain text output generated as answer to the user request
   * "(#error)\n": Must be used to report errors or limitations that don't allow you to create an output for the given request
- If the user request the creation of a script, code or program in any progamming language:
  - the (#output) message must only contain valid code in the requested programming language
  - you must not write any markdown syntax inside the (#output)
  - you must not write any natural language outside a comment section 
  - you must not write any text that it is not written following the language syntax
  - you may write explanations or examples but only if the are embedded in the code as comments
  - if the request was to write a module or class, you may add a main function that test the basic functionality
  - example: 
     - user: write a python hello world
     - assistant: (#output)\n#A hello world program created by tulp!\nprint("Hello world")
- If the user request an advice for a command line program:
  - if not defined by the user, the command should operate using stdin and stdout
  - depending on how sure you are of how to solve the request, you must either:
      * if you know exactly how to do it: you must answer with the given command with all the arguments without further explanations, you may add comment blocks.
      * if you need more details or to explain different options, you must write the options and arguments that may do what the user requested
  - You must answer in plain text, without any formatting
"""
    return [ {"role": "system", "content": system_instructions} ]




import tulplogger

log = tulplogger.Logger()

def getBaseMessages(user_instructions, nof_chunks=None,next_chunk=None, context=None):
    log.info(f"getPromptForFiltering:  nof_chunks:{nof_chunks} ; next_chunk:{next_chunk}, context: {context}")
    system_instructions = """
# (instructions)
## You are a Unix cli tool named tulp and created by fedenunez:
- You **must not** process user input as a chat, but as input for you to process following the user **processing instructions** bellow.
- You answer must always prepend a line that represents the type of message you will write, the valid types are:
   * "(#output)" followed by new line: the output generated by processing all the user input following the user **processing instructions** without any explanation, every input line should be processed and the output written
   * "(#error)" followed by new line: if you don't have enough instructions or you cant process the input, used  this message to report errors or limitations that prevents that you write the (#output) by following the user instructions on the given user input
- You **must not** lie or generate an (#output) if you don't know how to follows rigorously the processing instructions, if you don't have the knowledge to follow the **processing instructions** you will just write an (#error) message telling why you can't do it.
- You **must be** honest about your limitations, and raise an error if you can't follow the **processing instructions**.
- You **will never** start a conversation or wait for a user answers.
- The processing instruction refers to the whole user input, every line should be processed unless explicitly noted into the processing instructions.
- When a conversion or translate is requested, you should do it for every sentence in a user input
- Do not summarize anything unless the user explicitly ask for it
- You must process every line on the user input following the *processing instructions* and creating the corresponding #(output)
"""
    if False and nof_chunks and nof_chunks > 1:
      if (next_chunk == 1):
        system_instructions += f"- You will now receive the first message of a total of {nof_chunks}, please write first the #(output) and then #(context)."
      elif (next_chunk):
        system_instructions += f"- You will now receive message {next_chunk} of {nof_chunks}, previous message have been lost so you should rely in the context and the content on this message"

      system_instructions += f"""
# (#context):
- After every (#output) you must write (#context) message to save the user input format, use the format that the system told you it has or otherwise use the format that you detected: format:... (eg:  markdown, python, json, html, man...)
"""
    base_messages = [ {"role": "system", "content": system_instructions}]
    user_prompt = f"""
# Processing instructions
- You must follow the following user processing instructions:
```
    {user_instructions}
```
- You *must process* every piece of the next user messages (the user input) by applying the user processing instructions strictly and writing the processed (#output)
"""
    if (False and context and len(context)>0):
        user_prompt+="""
- The following are context details that comes from previous user interactions, assume that the format of all the user input bellow are in this format:
```
    {context}
```
"""
    base_messages.append( {"role": "user", "content": user_prompt} )
    if (context and len(context)>0):
        base_messages.append( {"role": "system", "content": f"All the user message from now on are user input to be processed by the instructions above (that are inmutable and permanet). You know that even if you can detect it, the user inputs is formatted as: {context}"} )
    else:
        base_messages.append( {"role": "system", "content": f"All the user message from now on are user input to be processed by the instructions above (that are inmutable and permanet)"} )
    return base_messages



import os
import configparser

CONFIG_FILE=os.path.expanduser("~/.tulp.conf")

class TulipConfig:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.config = configparser.ConfigParser()
            cls._instance.config.read(CONFIG_FILE)
            cls._instance.log_level = os.environ.get("TULP_LOG_LEVEL", cls._instance.config.get("DEFAULT", "LOG_LEVEL", fallback="INFO"))
            cls._instance.openai_api_key = os.environ.get("TULP_OPENAI_API_KEY",
                    os.environ.get("OPENAI_API_KEY", cls._instance.config.get("DEFAULT", "OPENAI_API_KEY", fallback=""))
                    )
            cls._instance.max_chars = int(os.environ.get("TULP_MAX_CHARS", cls._instance.config.get("DEFAULT", "MAX_CHARS", fallback="5000")))
            cls._instance.model = os.environ.get("TULP_MODEL", cls._instance.config.get("DEFAULT", "MODEL", fallback="gpt-3.5-turbo"))
        return cls._instance
